use std::{net::SocketAddr, sync::Arc};

use log::{error, info};
use quick_xml::{
    events::{attributes::Attributes, BytesDecl, BytesEnd, BytesStart, Event},
    Reader, Writer,
};
use tokio::{
    io::{AsyncWrite, AsyncWriteExt, BufReader, ReadHalf, WriteHalf},
    net::TcpStream,
};
use tokio_rustls::rustls;

use crate::{
    error::StreamError,
    feature::Feature,
    negotiator::{self, Step, TcpConnOrTLS},
    tag::{self, Tag},
    tls::stream,
};

pub(crate) enum StartTLSResult {
    Success(stream::TLSStream),
    Failure(StreamStart, StreamError),
    /// TLSFailure is returned when the TLS negotiation failed, and the TCP
    /// stream is to be dropped.
    TLSFailure(StreamError),
}

type Result<T> = std::result::Result<T, StreamError>;

const FEATURES: &'static [Feature] = &[Feature::start_tls(true)];

pub struct StreamStart {
    reader: Reader<BufReader<ReadHalf<TcpStream>>>,
    writer: Writer<WriteHalf<TcpStream>>,
    buffer: Vec<u8>,
    hostname: String,
    tls_config: Arc<rustls::ServerConfig>,
}

impl StreamStart {
    fn new(tcp_stream: TcpStream, hostname: String, tls_config: Arc<rustls::ServerConfig>) -> Self {
        let (read, write) = tokio::io::split(tcp_stream);
        let (reader, writer) = (
            Reader::from_reader(BufReader::new(read)),
            Writer::new(write),
        );

        Self {
            reader,
            writer,
            hostname,
            tls_config,
            buffer: vec![],
        }
    }

    async fn start_stream(self) {
        match self.negotiate_stream().await {
            StartTLSResult::Success(tls_stream) => tls_stream.start_stream().await,
            StartTLSResult::Failure(mut conn, err) => {
                if let Err(err2) = error(conn.writer.get_mut(), err).await {
                    error!("error writing error: {err2}");
                    return;
                } else {
                    info!("wrote error {err}")
                }

                if let Err(e) = conn.writer.get_mut().write_all(b"</stream:stream>").await {
                    error!("writing end to stream: {e}")
                }
                if let Err(e) = conn.writer.get_mut().shutdown().await {
                    error!("shutting down stream: {e}")
                }
            }
            StartTLSResult::TLSFailure(err) => {
                error!("TLS negotiation failure: {err}. Dropping connection.")
            }
        }
    }

    async fn negotiate_stream(mut self) -> StartTLSResult {
        let attrs = loop {
            let event = match self.reader.read_event_into_async(&mut self.buffer).await {
                Ok(event) => event,
                Err(err) => return StartTLSResult::Failure(self, err.into()),
            };
            match event {
                Event::Start(start) => {
                    if start.name().as_ref() == tag::STREAM_ELEMENT_NAME {
                        let attrs: StreamAttrs = match start.attributes().try_into() {
                            Ok(a) => a,
                            Err(err) => return StartTLSResult::Failure(self, err.into()),
                        };
                        if attrs.namespace != XMLNamespace::JabberClient {
                            return StartTLSResult::Failure(self, StreamError::InvalidNamespace);
                        }
                        break attrs;
                    } else {
                        info!("element: {:?}", start);
                    }
                }
                Event::End(_) => return StartTLSResult::Failure(self, StreamError::BadFormat),
                Event::Eof => return StartTLSResult::Failure(self, StreamError::BadFormat),
                _ => continue,
            }
        };
        info!("starting negotiation with: {attrs:?}");
        if let Err(err) = self
            .write_stream_header(StreamAttrs {
                from: attrs.to.clone(),
                to: attrs.from,
                namespace: XMLNamespace::JabberClient,
            })
            .await
        {
            return StartTLSResult::Failure(self, err);
        };
        if attrs.to != self.hostname {
            return StartTLSResult::Failure(self, StreamError::HostUnknown);
        }
        if let Err(err) = self.send_features().await {
            return StartTLSResult::Failure(self, err);
        };
        loop {
            let event = match self.reader.read_event_into_async(&mut self.buffer).await {
                Ok(event) => event,
                Err(err) => return StartTLSResult::Failure(self, err.into()),
            };
            match event {
                Event::Empty(empty) => match empty.name().as_ref() {
                    tag::STARTTLS => {
                        info!("starttls negotiation");
                        let hostname = self.hostname;
                        let tls_config = self.tls_config.clone();
                        return match negotiator::start_tls(
                            self.reader
                                .into_inner()
                                .into_inner()
                                .unsplit(self.writer.into_inner()),
                            empty,
                            self.tls_config.clone(),
                        )
                        .await
                        {
                            TcpConnOrTLS::TLS(conn) => {
                                StartTLSResult::Success(stream::TLSStream::new(conn, hostname))
                            }
                            TcpConnOrTLS::TCP(conn, err) => StartTLSResult::Failure(
                                StreamStart::new(conn, hostname, tls_config),
                                err,
                            ),
                            TcpConnOrTLS::Failed(err) => StartTLSResult::TLSFailure(err),
                        };
                    }
                    _ => return StartTLSResult::Failure(self, StreamError::UnsupportedFeature),
                },
                Event::End(_) => return StartTLSResult::Failure(self, StreamError::BadFormat),
                Event::Eof => return StartTLSResult::Failure(self, StreamError::BadFormat),
                _ => continue,
            }
        }

        StartTLSResult::Failure(self, StreamError::InternalServerError)
    }
    async fn write_stream_header(&mut self, req: StreamAttrs) -> Result<()> {
        self.writer
            .write_event_async(Event::Decl(BytesDecl::new("1.0", Some("utf-8"), None)))
            .await?;
        self.writer
            .write_event_async(Event::Start(
                BytesStart::new("stream:stream").with_attributes(vec![
                    ("from", req.from.as_str()),
                    ("to", req.to.as_str()),
                    ("xmlns:stream", "http://etherx.jabber.org/streams"),
                    ("xml:lang", "en"),
                    ("version", "1.0"),
                ]),
            ))
            .await?;

        Ok(())
    }

    async fn send_features(&mut self) -> Result<()> {
        self.writer
            .write_event_async(Event::Start(BytesStart::new(tag::FEATURE)))
            .await?;
        for feature in FEATURES.into_iter() {
            feature.write_tag(self.writer.get_mut()).await?;
        }
        self.writer
            .write_event_async(Event::End(BytesEnd::new(tag::FEATURE)))
            .await?;
        Ok(())
    }
}

pub fn spawn(
    hostname: String,
    (mut stream, _): (TcpStream, SocketAddr),
    tls_config: Arc<rustls::ServerConfig>,
) {
    tokio::spawn(async move {
        StreamStart::new(stream, hostname, tls_config)
            .start_stream()
            .await;
    });
}

async fn error<W: AsyncWrite + Unpin>(writer: W, err: StreamError) -> Result<()> {
    let mut writer = Writer::new(writer);
    let err = err.to_string();
    writer
        .write_event_async(Event::Start(BytesStart::new(tag::ERROR_ELEMENT)))
        .await?;
    writer
        .write_event_async(Event::Start(
            BytesStart::new(&err)
                .with_attributes(vec![("xmlns", "urn:ietf:params:xml:ns:xmpp-streams")]),
        ))
        .await?;
    writer
        .write_event_async(Event::End(BytesEnd::new(&err)))
        .await?;
    writer
        .write_event_async(Event::End(BytesEnd::new(tag::ERROR_ELEMENT)))
        .await?;
    Ok(())
}

#[derive(Debug, Clone)]
struct StreamAttrs {
    from: String,
    to: String,
    namespace: XMLNamespace,
}

impl TryFrom<Attributes<'_>> for StreamAttrs {
    type Error = StreamError;

    fn try_from(value: Attributes<'_>) -> std::result::Result<Self, Self::Error> {
        let mut from: Option<String> = None;
        let mut to: Option<String> = None;
        let mut ns: Option<XMLNamespace> = None;
        for v in value {
            let v = v?;
            match v.key.local_name().into_inner() {
                b"from" => {
                    from = Some(String::from_utf8(v.value.to_vec())?);
                }
                b"to" => {
                    to = Some(String::from_utf8(v.value.to_vec())?);
                }
                b"xmlns" => match v.value.to_vec().as_slice() {
                    b"jabber:client" => {
                        ns = Some(XMLNamespace::JabberClient);
                    }
                    _ => return Err(StreamError::InvalidNamespace),
                },
                other => {
                    info!(
                        "ignoring key {}",
                        String::from_utf8(other.to_vec()).unwrap_or_default()
                    );
                }
            }
        }
        Ok(StreamAttrs {
            from: from.ok_or(StreamError::InvalidFrom)?,
            to: to.ok_or(StreamError::HostUnknown)?,
            namespace: ns.ok_or(StreamError::BadNamespacePrefix)?,
        })
    }
}

#[derive(Debug, Clone, Copy, PartialEq, Eq)]
enum XMLNamespace {
    JabberClient,
}
